﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace Ustawienia
{
    public class SettingsManager<T>
        where T : class, new()
    {
        //private T pola = default(T);

        public Dictionary<string, string> Values { get; private set; }

        public SettingsManager(T settingsObject, IFormatProvider formatProvider = null)
        {
            if (formatProvider == null) formatProvider = CultureInfo.InvariantCulture;
            Values = extractValues(settingsObject, formatProvider);
        }

        private static bool isProper(FieldInfo field)
        {
            bool result = !field.IsNotSerialized && !field.IsStatic;
            return result;
        }

        private static bool isProper(PropertyInfo property)
        {
            bool result = property.CanWrite;
            return result;
        }

        #region Wydobywanie informacji o elementach składowych, bez rekurencji
        public static List<string> ExtractMembersNames()
        {
            List<string> memberNames = new List<string>();

            //reflection
            Type type = typeof(T);

            FieldInfo[] fields = type.GetFields();
            foreach (FieldInfo field in fields)
                if (isProper(field))
                    memberNames.Add(field.Name);

            PropertyInfo[] properties = type.GetProperties();
            foreach (PropertyInfo property in properties)
                if (isProper(property))
                    memberNames.Add(property.Name);

            return memberNames;
        }

        private static Dictionary<string, string> extractValues(T settingsObject, IFormatProvider formatProvider)
        {
            if (settingsObject == null) throw new SettingsException("Obiekt nie może być pusty (null)");
            else
            {
                Dictionary<string, string> values = new Dictionary<string, string>();

                Type type = typeof(T);

                FieldInfo[] fields = type.GetFields();
                foreach(FieldInfo field in fields)
                {
                    if(isProper(field))
                    {
                        object value = field.GetValue(settingsObject);
                        //values.Add(field.Name, value);
                        values.Add(field.Name, Convert.ToString(value, formatProvider));
                    }
                }

                PropertyInfo[] properties = type.GetProperties();
                foreach (PropertyInfo property in properties)
                {
                    if (isProper(property))
                    {
                        object value = property.GetValue(settingsObject);
                        //values.Add(property.Name, value);
                        values.Add(property.Name, Convert.ToString(value, formatProvider));
                    }
                }

                return values;
            }
        }

        private static T buildObjectFromValues(Dictionary<string, string> values, IFormatProvider formatProvider)
        {
            T settingObject = new T();

            FieldInfo[] fields = typeof(T).GetFields();
            foreach (FieldInfo field in fields)
            {
                if (isProper(field))
                {
                    object value = null;
                    if (!values.ContainsKey(field.Name)) value = Activator.CreateInstance(field.FieldType);
                    else value = Convert.ChangeType(values[field.Name], field.FieldType, formatProvider);
                    field.SetValue(settingObject, value);
                }
            }

            PropertyInfo[] properties = typeof(T).GetProperties();
            foreach (PropertyInfo property in properties)
            {
                if (isProper(property))
                {
                    object value = null;
                    if (!values.ContainsKey(property.Name)) value = Activator.CreateInstance(property.PropertyType);
                    else value = Convert.ChangeType(values[property.Name], property.PropertyType, formatProvider);
                    property.SetValue(settingObject, value);
                }
            }

            return settingObject;
        }
        #endregion

        #region Obsługa plików XML
        private const string settingElementName = "Setting";
        private const string nameAttributeName = "Name";

        public void SaveToXml(string filePath)
        {
            XDocument xml = new XDocument(
                new XDeclaration("1.0", "utf-8", "yes"),
                new XElement("Settings",
                    from KeyValuePair<string, string> value in Values
                    orderby value.Key
                    select new XElement(settingElementName,
                        new XAttribute(nameAttributeName, value.Key),
                        value.Value)));
            xml.Save(filePath);
        }

        private static Dictionary<string, string> loadValuesFromXml(string path)
        {
            Dictionary<string, string> values = new Dictionary<string, string>();
            XDocument xml = XDocument.Load(path);
            foreach(XElement element in xml.Descendants(settingElementName))
            {
                string name = element.Attribute(nameAttributeName).Value;
                string value = element.Value;
                values.Add(name, value);
            }
            return values;
        }

        public static SettingsManager<T> LoadFromXml(string path, IFormatProvider formatProvider = null)            
        {
            if (!System.IO.File.Exists(path)) throw new Exception("Plik " + path + " nie istnieje");
            if (formatProvider == null) formatProvider = CultureInfo.InvariantCulture;
            Dictionary<string, string> values = loadValuesFromXml(path);
            T settingsObject = buildObjectFromValues(values, formatProvider);
            return new SettingsManager<T>(settingsObject, formatProvider);
        }
        #endregion
    }

    public class SettingsException : Exception
    {
        public SettingsException(string message, Exception innerException = null)
            :base(message, innerException)
        {


        }
    }
}
